/*
 *  acpiconf.c - ACPI based kernel configuration
 *
 *  Copyright (C) 2000-2001 Intel Corp.
 *  Copyright (C) 2000-2001 J.I. Lee <Jung-Ik.Lee@intel.com>
 *
 *  Revision History:
 *	 9/15/2000 J.I.
 *		Major revision:	for new ACPI initialization requirements
 *	11/15/2000 J.I.
 *		Major revision:	ACPI 2.0 tables support
 *	04/23/2001 J.I.
 *		Rewrote functions to support multiple _PRTs of child P2Ps
 *		under root pci bus
 */


#include <linux/config.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/pci.h>
#include <asm/system.h>
#include <asm/iosapic.h>
#include <asm/efi.h>
#include <asm/acpikcfg.h>
#include "acpi.h"
#include "osconf.h"
#include "acpiconf.h"


static	int	acpi_cf_initialized __initdata = 0;

ACPI_STATUS	__init
acpi_cf_init (
	void * rsdp
	)
{
	ACPI_STATUS	status;

	acpi_os_bind_osd(ACPI_CF_PHASE_BOOTTIME);

	status = acpi_initialize_subsystem ();
	if (ACPI_FAILURE(status)) {
		printk ("Acpi cfg:initialize_subsystem error=0x%x\n", status);
		return status;
	}
	dprintk(("Acpi cfg:initializ_subsysteme pass\n"));

	status = acpi_load_tables ((u64)rsdp);
	if (ACPI_FAILURE(status)) {
		printk ("Acpi cfg:load firmware tables error=0x%x\n", status);
		acpi_terminate();
		return status;
	}
	dprintk(("Acpi cfg:load firmware tables pass\n"));

	status = acpi_enable_subsystem (ACPI_FULL_INITIALIZATION);
	if (ACPI_FAILURE(status)) {
		printk ("Acpi cfg:enable_subsystem error=0x%x\n", status);
		acpi_terminate();
		return status;
	}
	dprintk(("Acpi cfg:enable_subsystem pass\n"));

	acpi_cf_initialized++;

	return AE_OK;
}


ACPI_STATUS	__init
acpi_cf_terminate ( void )
{
	ACPI_STATUS	status;

	if (! ACPI_CF_INITIALIZED())
		return AE_ERROR;

	status = acpi_disable ();
	if (ACPI_FAILURE(status)) {
		printk ("Acpi cfg:disable fail=0x%x\n", status);
		/* fall thru...*/
	}

	status = acpi_terminate ();
	if (ACPI_FAILURE(status)) {
		printk ("Acpi cfg:acpi terminate error=0x%x\n", status);
		/* fall thru...*/
	}

	acpi_cf_cleanup();
	acpi_os_bind_osd(ACPI_CF_PHASE_RUNTIME);

	acpi_cf_initialized--;

	return status;
}


ACPI_STATUS	__init
acpi_cf_get_pci_vectors (
	struct pci_vector_struct	**vectors,
	int				*num_pci_vectors
	)
{
	ACPI_STATUS	status;
	void		*prts;

	if (! ACPI_CF_INITIALIZED()) {
		status = acpi_cf_init((void *)efi.acpi);
		if (ACPI_FAILURE (status))
			return status;
	}

	*vectors = NULL;
	*num_pci_vectors = 0;

	status = acpi_cf_get_prt (&prts);
	if (ACPI_FAILURE (status)) {
		printk("Acpi cfg: get prt fail\n");
		return status;
	}

	status = acpi_cf_convert_prt_to_vectors (prts, vectors, num_pci_vectors);
#ifdef	CONFIG_ACPI_KERNEL_CONFIG_DEBUG
	if (ACPI_SUCCESS(status)) {
		acpi_cf_print_pci_vectors (*vectors, *num_pci_vectors);
	}
#endif
	printk("Acpi cfg: get PCI interrupt vectors %s\n",
		(ACPI_SUCCESS(status))?"pass":"fail");

	return status;
}


static	PCI_ROUTING_TABLE	*pci_routing_tables[PCI_MAX_BUS] __initdata = {NULL};


typedef struct _acpi_rpb {
	NATIVE_UINT	rpb_busnum;
	NATIVE_UINT	lastbusnum;
	ACPI_HANDLE	rpb_handle;
} acpi_rpb_t;


static ACPI_STATUS	__init
acpi_cf_evaluate_method (
	ACPI_HANDLE	handle,
	UINT8		*method_name,
	NATIVE_UINT	*nuint
	)
{
	UINT32			tnuint = 0;
	ACPI_STATUS		status;

	ACPI_BUFFER		ret_buf;
	ACPI_OBJECT		*ext_obj;
	UINT8			buf[PATHNAME_MAX];


	ret_buf.length  = PATHNAME_MAX;
	ret_buf.pointer = (void *) buf;

	status = acpi_evaluate_object(handle, method_name, NULL, &ret_buf);
	if (ACPI_FAILURE(status)) {
		if (status == AE_NOT_FOUND) {
			printk("Acpi cfg: no %s found\n", method_name);
		} else {
			printk("Acpi cfg: %s fail=0x%x\n", method_name, status);
		}
	} else {
		ext_obj = (ACPI_OBJECT *) ret_buf.pointer;

		switch (ext_obj->type) {
		case ACPI_TYPE_INTEGER:
			tnuint = (NATIVE_UINT) ext_obj->integer.value;
			break;
		default:
			printk("Acpi cfg: %s obj type incorrect\n", method_name);
			status = AE_TYPE;
			break;
		}
	}

	*nuint = tnuint;
	return (status);
}


static ACPI_STATUS	__init
acpi_cf_evaluate_PRT (
	ACPI_HANDLE		handle,
	PCI_ROUTING_TABLE	**prt
	)
{
	ACPI_BUFFER		acpi_buffer;
	ACPI_STATUS		status;

	acpi_buffer.length  = 0;
	acpi_buffer.pointer = NULL;

	status = acpi_get_irq_routing_table (handle, &acpi_buffer);

	switch (status) {
	case AE_BUFFER_OVERFLOW:
		dprintk(("Acpi cfg: _PRT found. need %d bytes\n",
			acpi_buffer.length));
		break;		/* found */
	default:
		printk("Acpi cfg: _PRT fail=0x%x\n", status);
	case AE_NOT_FOUND:
		return status;
	}

	*prt = (PCI_ROUTING_TABLE *) acpi_os_callocate (acpi_buffer.length);
	if (!*prt) {
		printk("Acpi cfg: callocate %d bytes for _PRT fail\n",
			acpi_buffer.length);
		return AE_NO_MEMORY;
	}
	acpi_buffer.pointer = (void *) *prt;

	status = acpi_get_irq_routing_table (handle, &acpi_buffer);
	if (ACPI_FAILURE(status)) {
		printk("Acpi cfg: _PRT fail=0x%x.\n", status);
		acpi_os_free(prt);
	}

	return status;
}

static ACPI_STATUS	__init
acpi_cf_get_root_pci_callback (
	ACPI_HANDLE		handle,
	UINT32			Level,
	void			*context,
	void			**retval
	)
{
	NATIVE_UINT		busnum = 0;
	ACPI_STATUS		status;
	acpi_rpb_t		rpb;
	PCI_ROUTING_TABLE	*prt;

	UINT8			path_name[PATHNAME_MAX];

#ifdef	CONFIG_ACPI_KERNEL_CONFIG_DEBUG
	ACPI_BUFFER		ret_buf;

	ret_buf.length  = PATHNAME_MAX;
	ret_buf.pointer = (void *) path_name;

	status = acpi_get_name(handle, ACPI_FULL_PATHNAME, &ret_buf);
#else
	memset(path_name, 0, sizeof (path_name));
#endif

	/*
	 * get bus number of this pci root bridge
	 */
	status = acpi_cf_evaluate_method(handle, METHOD_NAME__BBN, &busnum);
	if (ACPI_FAILURE(status)) {
		printk("Acpi cfg:%s evaluate _BBN fail=0x%x\n",
			path_name, status);
		return (status);
	}
	printk("Acpi cfg:%s ROOT PCI bus %ld\n", path_name, busnum);

	/*
	 * evaluate root pci bridge's _CRS for Bus number range for child P2P
	 * (bus min/max/len) - not yet.
	 */

	/*
	 * get immediate _PRT of this root pci bridge if any
	 */
	status = acpi_cf_evaluate_PRT (handle, &prt);
	switch(status) {
	case AE_NOT_FOUND:
		break;
	default:
		if (ACPI_FAILURE(status)) {
			printk("Acpi cfg:%s _PRT fail=0x%x\n",
				path_name, status);
			return status;
		}
		dprintk(("Acpi cfg:%s bus %ld got _PRT\n", path_name, busnum));
		acpi_cf_add_to_pci_routing_tables (busnum, prt);
		break;
	}


	/*
	 * walk down this root pci bridge to get _PRTs if any
	 */
	rpb.rpb_busnum = rpb.lastbusnum = busnum;
	rpb.rpb_handle = handle;
	status = acpi_walk_namespace ( ACPI_TYPE_DEVICE,
					handle,
					ACPI_UINT32_MAX,
					acpi_cf_get_prt_callback,
					&rpb,
					NULL );
	if (ACPI_FAILURE(status))
		printk("Acpi cfg:%s walk namespace for _PRT error=0x%x\n",
			path_name, status);

	return (status);
}


/*
 * handle _PRTs of immediate P2Ps of root pci. 
 */
static ACPI_STATUS	__init
acpi_cf_associate_prt_to_bus (
	ACPI_HANDLE		handle,
	acpi_rpb_t		*rpb,
	NATIVE_UINT		*retbusnum,
	NATIVE_UINT		depth
	)
{
	ACPI_STATUS		status;
	UINT32			segbus;
	NATIVE_UINT		devfn;
	UINT8			bn;

	UINT8			path_name[PATHNAME_MAX];

#ifdef	CONFIG_ACPI_KERNEL_CONFIG_DEBUG
	ACPI_BUFFER		ret_buf;

	ret_buf.length  = PATHNAME_MAX;
	ret_buf.pointer = (void *) path_name;

	status = acpi_get_name(handle, ACPI_FULL_PATHNAME, &ret_buf);
#else
	memset(path_name, 0, sizeof (path_name));
#endif

	/*
	 * get devfn from _ADR
	 */
	status = acpi_cf_evaluate_method(handle, METHOD_NAME__ADR, &devfn);
	if (ACPI_FAILURE(status)) {
		*retbusnum = rpb->rpb_busnum + 1;
		printk("Acpi cfg:%s _ADR fail=0x%x. Set busnum to %ld\n",
			path_name, status, *retbusnum);
		return AE_OK;
	}
	dprintk(("Acpi cfg:%s _ADR =0x%x\n", path_name, (UINT32)devfn));


	/*
	 * access pci config space for bus number
	 *	segbus = from rpb, devfn = from _ADR
	 */
	segbus = (UINT32)(rpb->rpb_busnum & 0xFFFFFFFF);

	status = acpi_os_read_pci_cfg_byte(segbus,
				(UINT32)devfn, PCI_PRIMARY_BUS, &bn);
	if (ACPI_FAILURE(status)) {
		*retbusnum = rpb->rpb_busnum + 1;
		printk("Acpi cfg:%s pci read fail=0x%x. b:df:a=%x:%x:%x\n",
			path_name, status, segbus, (UINT32)devfn,
			PCI_PRIMARY_BUS);
		printk("Acpi cfg:%s Set busnum to %ld\n",
			path_name, *retbusnum);
		return AE_OK;
	}
	dprintk(("Acpi cfg:%s pribus %d\n", path_name, bn));


	status = acpi_os_read_pci_cfg_byte(segbus,
				(UINT32)devfn, PCI_SECONDARY_BUS, &bn);
	if (ACPI_FAILURE(status)) {
		*retbusnum = rpb->rpb_busnum + 1;
		printk("Acpi cfg:%s pci read fail=0x%x. b:df:a=%x:%x:%x\n",
			path_name, status, segbus, (UINT32)devfn,
			PCI_SECONDARY_BUS);
		printk("Acpi cfg:%s Set busnum to %ld\n",
			path_name, *retbusnum);
		return AE_OK;
	}
	dprintk(("Acpi cfg:%s busnum %d\n", path_name, bn));

	*retbusnum = (NATIVE_UINT)bn;
	return AE_OK;
}


static ACPI_STATUS	__init
acpi_cf_get_prt (
	void		**prts
	)
{
	ACPI_STATUS	status;

	status = acpi_get_devices ( PCI_ROOT_HID_STRING,
				acpi_cf_get_root_pci_callback,
				NULL,
				NULL );

	if (ACPI_FAILURE(status)) {
		printk("Acpi cfg:get_device PCI ROOT HID error=0x%x\n", status);
	}

	*prts = (void *)pci_routing_tables;

	return status;
}

static ACPI_STATUS	__init
acpi_cf_get_prt_callback (
	ACPI_HANDLE		handle,
	UINT32			Level,
	void			*context,
	void			**retval
	)
{
	PCI_ROUTING_TABLE	*prt;
	NATIVE_UINT		busnum = 0;
	NATIVE_UINT		temp = 0x0F;
	ACPI_STATUS		status;

	UINT8			path_name[PATHNAME_MAX];

#ifdef	CONFIG_ACPI_KERNEL_CONFIG_DEBUG
	ACPI_BUFFER		ret_buf;

	ret_buf.length  = PATHNAME_MAX;
	ret_buf.pointer = (void *) path_name;

	status = acpi_get_name(handle, ACPI_FULL_PATHNAME, &ret_buf);
#else
	memset(path_name, 0, sizeof (path_name));
#endif

	status = acpi_cf_evaluate_PRT (handle, &prt);
	switch(status) {
	case AE_NOT_FOUND:
		return AE_OK;
	default:
		if (ACPI_FAILURE(status)) {
			printk("Acpi cfg:%s _PRT fail=0x%x\n",
				path_name, status);
			return status;
		}
	}

	/*
	 * evaluate _STA in case this device does not exist
	 */
	status = acpi_cf_evaluate_method(handle, METHOD_NAME__STA, &temp);
	switch(status) {
	case AE_NOT_FOUND:
		break;
	default:
		if (ACPI_FAILURE(status)) {
			printk("Acpi cfg:%s _STA fail=0x%x\n",
				path_name, status);
			return status;
		}
		if (!(temp & ACPI_STA_DEVICE_PRESENT)) {
			dprintk(("Acpi cfg:%s not exist. _PRT discarded\n",
				path_name));
			acpi_os_free(prt);
			return AE_OK;
		}
		break;
	}

	/*
	 * associate a bus number to this _PRT since 
	 * this _PRT is not on root pci bridge 
	 */
	acpi_cf_associate_prt_to_bus(handle, context, &busnum, 0);

	printk("Acpi cfg:%s busnum %ld got _PRT\n", path_name, busnum);
	acpi_cf_add_to_pci_routing_tables (busnum, prt);

	return AE_OK;
}


static void	__init
acpi_cf_add_to_pci_routing_tables (
	NATIVE_UINT		busnum,
	PCI_ROUTING_TABLE	*prt
	)
{
	if ( busnum >= PCI_MAX_BUS ) {
		printk("Acpi cfg:invalid pci bus number %ld\n", busnum);
		acpi_os_free(prt);
		return;
	}

	if (pci_routing_tables[busnum]) {
		printk("Acpi cfg:duplicate PRT for pci bus %ld. overiding...\n", busnum);
		acpi_os_free(pci_routing_tables[busnum]);
	}

	pci_routing_tables[busnum] = prt;
}


#define	DUMPVECTOR(pv)	printk("PCI bus=0x%x id=0x%x pin=0x%x irq=0x%x\n", pv->bus, pv->pci_id, pv->pin, pv->irq);

static ACPI_STATUS	__init
acpi_cf_convert_prt_to_vectors (
	void				*prts,
	struct pci_vector_struct	**vectors,
	int				*num_pci_vectors
	)
{
	struct pci_vector_struct	*pvec;
	PCI_ROUTING_TABLE		**pprts, *prt, *prtf;
	int				nvec = 0;
	int				i;


	pprts = (PCI_ROUTING_TABLE **)prts;

	for ( i = 0; i < PCI_MAX_BUS; i++) {
		prt = *pprts++;
		if (prt) {
			for ( ; prt->length > 0; nvec++) {
				prt = (PCI_ROUTING_TABLE *) ((NATIVE_UINT)prt + (NATIVE_UINT)prt->length);
			}
		}
	}

	*num_pci_vectors = nvec;
	*vectors = acpi_os_callocate (sizeof(struct pci_vector_struct) * nvec);
	if (!*vectors) {
		printk("Acpi cfg: callocate for pci_vector error\n");
		return AE_NO_MEMORY;
	}

	pvec = *vectors;
	pprts = (PCI_ROUTING_TABLE **)prts;

	for ( i = 0; i < PCI_MAX_BUS; i++) {
		prt = prtf = *pprts++;
		if (prt) {
			for ( ; prt->length > 0; pvec++) {
				pvec->bus	= (UINT16)i;
				pvec->pci_id	= prt->address;
				pvec->pin	= (UINT8)prt->pin;
				pvec->irq	= (UINT8)prt->source_index;

				prt = (PCI_ROUTING_TABLE *) ((NATIVE_UINT)prt + (NATIVE_UINT)prt->length);
			}
			acpi_os_free((void *)prtf);
		}
	}

	return AE_OK;
}


void	__init
acpi_cf_cleanup ( void )
{
	/* nothing to free, pci_vectors are used by the kernel */
}


#ifdef	CONFIG_ACPI_KERNEL_CONFIG_DEBUG
void	__init
acpi_cf_print_pci_vectors (
	struct pci_vector_struct	*vectors,
	int				num_pci_vectors
	)
{
	struct pci_vector_struct	*pvec;
	int				i;

	printk("number of PCI interrupt vectors = %d\n", num_pci_vectors);

	pvec = vectors;
	for (i = 0; i < num_pci_vectors; i++) {
		DUMPVECTOR(pvec);
		pvec++;
	}
}
#endif
